/*
 * Datei: feld.c
 * Autor: Holger Klawitter
 * Datum: 13.8.1998
 *
 * Changes:
 * 10.6.1999  Flip added
 * 27.9.1999  +Empty()
 */

#include "feld.h"
#include "hmaki.h"
#include "hofDB.h"
#include "hof.h"
#include <PalmOS.h>

/**********************************************************************
 * Public variables and constants.
 **********************************************************************/

Byte *feld;
Boolean hasFeld = false;
Boolean cheater = false;

Boolean undoActive = false;

/**********************************************************************
 * Private variables and constants.
 **********************************************************************/

Byte *Nfeld;
/* Temporary storage for new field data.
 */

typedef struct {
	Byte *data;
	Int32 score;
} UndoData;

#define MAXUNDO 5

UndoData *undo[MAXUNDO] = { 0,0,0,0,0 };
/* Last field for undo-operations.
 */

UInt undoCount;
/* index where the next undo would be written (if == MAXUNDO)
 * may be >= MAXUNDO, which means that next undo will have to
 * shift
 */

#if 1
CustomPatternType patterns[10] = {
  { 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00 }, /* white  */
  { 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff }, /* black  */
  { 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55 }, /* gray   */
  { 0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f }, /* sqares */
  { 0xcc, 0xcc, 0x33, 0x33, 0xcc, 0xcc, 0x33, 0x33 }, /* small sqares */
  { 0x88, 0x44, 0x22, 0x11, 0x88, 0x44, 0x22, 0x11 }, /* lr diagonals */
  { 0x11, 0x22, 0x44, 0x88, 0x11, 0x22, 0x44, 0x88 }, /* rl diagonals */
  { 0xaa, 0x44, 0xaa, 0x11, 0xaa, 0x44, 0xaa, 0x11 }, /* net */
  { 0xff, 0x00, 0xff, 0x00, 0xff, 0x00, 0xff, 0x00 }, /* hline  */
  { 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa }, /* vline  */
};
#else
// gcc seems to have a bug when using typedef'd arrays
// (internal error--insn does not satisfy its constraints)
Char patterns[80] = {
   0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, 0x00, /* white  */
   0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, 0xff, /* black  */
   0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, 0xaa, 0x55, /* gray   */
   0xf0, 0xf0, 0xf0, 0xf0, 0x0f, 0x0f, 0x0f, 0x0f, /* sqares */
   0xcc, 0xcc, 0x33, 0x33, 0xcc, 0xcc, 0x33, 0x33, /* small sqares */
   0x88, 0x44, 0x22, 0x11, 0x88, 0x44, 0x22, 0x11, /* lr diagonals */
   0x11, 0x22, 0x44, 0x88, 0x11, 0x22, 0x44, 0x88, /* rl diagonals */
   0xaa, 0x44, 0xaa, 0x11, 0xaa, 0x44, 0xaa, 0x11, /* net */
   0xff, 0x00, 0xff, 0x00, 0xff, 0x00, 0xff, 0x00, /* hline  */
   0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, 0xaa, /* vline  */
};
#endif

RGBColorType patterncolors[11] = {
	{ 0, 0xff, 0xff, 0xff },
	{ 1, 0xff, 0x00, 0x00 },
	{ 2, 0x00, 0xd0, 0x00 },
	{ 3, 0x00, 0x00, 0xd0 },
	{ 4, 0xff, 0xff, 0x00 },
	{ 5, 0x00, 0xff, 0xff },
	{ 6, 0xff, 0x40, 0xff },
	{ 7, 0xff, 0xa0, 0xa0 },
	{ 8, 0xa0, 0xff, 0xa0 },
	{ 9, 0xa0, 0xa0, 0xff },
	{10, 0x00, 0x00, 0x00 }, /* black */
};

/* Patterns indicating different colors.
 */

Boolean *marks;
/* Storage for marked area.
 */

Word hsize;
Word vsize;
/* Height and width of a single square.
 */

Long score;
/* Current score */

/**********************************************************************
 * Local Variables
 * Set in Alloc
 **********************************************************************/

Word height, width, colors;
PlayModes mode;

/**********************************************************************
 * Private Functions prototypes
 **********************************************************************/
void doSetScore( Long i );

/**********************************************************************
 * Functions
 **********************************************************************/
void Create(Word seed) {
  Int i,size;
  Word tmp = SysRandom(seed);
  Alloc();
  for( i=(size=height*width)-1; i>=0; i-- ) {
    feld[i] = (SysRandom(0)>>4) % colors + 1;
    /* The lowermost bits are usually not that random, so we use others. */
  }
  switch( mode ) {
  case modeHighScore:
  case modeLowScore:
    score = 0;
    break;
  case modeCleanBoard:
  case modeLeaveBoard:
    score = size;
  }
  SysRandom(tmp);
}

void Alloc() {
  Int i;

  height = currentHof.height;
  width = currentHof.width;
  colors = currentHof.colors;
  mode = currentHof.mode;

  if( hasFeld ) {
		Dealloc();
  }

  feld = MemPtrNew( height*width*sizeof(Byte) );
  Nfeld = MemPtrNew( height*width*sizeof(Byte) );
  marks = MemPtrNew( height*width*sizeof(Boolean) );
  hasFeld = true;
  undoCount = 0;
	cheater = false;
	undoActive = false;
  hsize = 160 / width;
  vsize = 140 / height;
  for( i=height*width-1; i>=0; i-- ) {
    Nfeld[i] = 8; /* for debugging, Nfeld should always be overwritten */
    marks[i] = false;
  }
}

void Dealloc() {
	Int i;

	MemPtrFree( feld );
	MemPtrFree( Nfeld );
	MemPtrFree( marks );
	for( i=0; i<MAXUNDO; i++ ) {
		if( undo[i] ) {
			MemPtrFree( undo[i]->data );
			MemPtrFree( undo[i] );
			undo[i] = NULL;
		}
	}
}

void doSetScore( Long val ) {
  VoidHand hand;
  CharPtr text;
  FieldPtr fld;

  hand = (VoidHand)FldGetTextHandle( fld=GetObjectPtr( MakiScore ) );
  if( !hand ) {
		hand = MemHandleNew(10);
    //FldSetTextHandle( fld, hand );
  }
  text = MemHandleLock( hand );
  StrIToA( text, val );
  MemHandleUnlock( hand );
  FldSetTextHandle( fld, (VoidPtr)hand );
  FldDrawField( fld );
}

void doPaint( SWord x, SWord y ) {
  RectangleType r = { { x*hsize, 160-(y+1)*vsize }, { hsize-1, vsize-1 } };
	RGBColorType rgb;
	Int col = feld[x*height+y];

	if( colorDevice && !bwMode ) {
		rgb = patterncolors[feld[x*height+y]];
		WinSetForeColor( WinRGBToIndex(&rgb) );
		WinPaintRectangle( &r, 0 );
	} else {
	  if( osVersion >= 0x35 ) WinSetPatternType( customPattern );
    WinSetPattern( (VoidPtr)(&patterns[col]) );
		WinFillRectangle( &r, 0 );
	}
}

void doSetMark( SWord x , SWord y ) {
  RectangleType r = { { x*hsize, 160-(y+1)*vsize }, { hsize-1, vsize-1 } };
	RGBColorType black;
	if( colorDevice && !bwMode ) {
		black = patterncolors[10];
		WinSetForeColor( WinRGBToIndex( &black ) );
	}
  WinDrawRectangleFrame( 1, &r );
}

void doClearMark( SWord x , SWord y ) {
  RectangleType r = { { x*hsize, 160-(y+1)*vsize }, { hsize-1, vsize-1 } };
  WinEraseRectangleFrame( 1, &r );
}

void Paint() {
  Int i,j;
  RectangleType r = { { 0, 19 }, { 160, 141 } };

	if( colorDevice && !bwMode ) {
		WinSetForeColor( 0 );
		WinPaintRectangle( &r, 0 );
		// force solid fill
	  if( osVersion >= 0x35 ) WinSetPatternType( customPattern );
    WinSetPattern( (VoidPtr)(&patterns[1]) );
	} else {
		if( osVersion >= 0x35 ) WinSetPatternType( customPattern );
		WinSetPattern( (const CustomPatternType*)(patterns) );
		WinFillRectangle( &r, 0 );
	}
  for( i=0; i<width; i++ ) {
    for( j=0; j<height; j++ ) {
      doPaint( i, j );
    }
  }
}

Boolean isLegal( Int x, Int y ) {
  return x>=0 && x<width && y>=0 && y<height;
}

Boolean isMarked( Int x, Int y ) {
  return isLegal(x,y) && marks[x*height+y];
}

Boolean Mark( Int x, Int y ) {
  Boolean hasChanged;
  Int i,j,col,count,tmp;

  /* Nothing to mark */
  if( !isLegal(x,y) ) {
    return false;
  }

  /* The user marked the field the second time. */
  if( feld[x*height+y] && marks[x*height+y] ) {
    return true;
  }

  UnMark();

  if( feld[x*height+y]==0 ) {
    return false;
  }

  /* Flood fill algorithm to determine the number of fields.
   */
  marks[x*height+y] = true;
  count = 1;
  do {
    hasChanged = false;
    for( i=0; i<width; i++ ) {
      col = i*height;
      for( j=0; j<height; j++ ) {
				tmp = feld[col+j];
				if( !marks[col+j] ) {
					if(
						 (i>0 && marks[col-height+j] && feld[col-height+j]==tmp) ||
						 (i<width-1 && marks[col+height+j] && feld[col+height+j]==tmp) ||
						 (j>0 && marks[col+j-1] && feld[col+j-1]==tmp) ||
						 (j<height-1 && marks[col+j+1] && feld[col+j+1]==tmp)
						 ) {
						marks[col+j] = true;
						count++;
						hasChanged = true;
					}
				} /* if */
      } /* for j */
    } /* for i */
  } while( hasChanged );

  /* Special case, single fields can't be marked.
   */
  if( count==1 ) {
    marks[x*height+y] = false;
    return false;
  }

  /* Mark the fields.
   */
  for( i=0; i<width*height; i++ ) {
    if( marks[i] ) {
      doSetMark( i/height, i%height );
    }
  }
  return false;
}

void undoShift() {
	Int i,size;
	UndoData *u;

	if( undoCount >= MAXUNDO ) {
		undoCount--;
		MemPtrFree( undo[0]->data );
		MemPtrFree( undo[0] );
		for( i=1; i<MAXUNDO; i++ ) {
			undo[i-1] = undo[i];
		}
	}
	
	undo[undoCount] = u = MemPtrNew( sizeof(UndoData) );
	u->score = score;
	u->data = MemPtrNew( size = sizeof(Byte)*height*width );
	MemMove( u->data, feld, size );

	undoActive=false;
	undoCount++;
}

Boolean Drop() {
  Int i,j,k,col,tmp, count = 0, occupied = 0;

	undoShift();

  for( i=0; i<width; i++ ) {
    for( j=0, k=0; j<height; j++ ) {
      tmp = Nfeld[i*height+k] = feld[i*height+j]; /* full copy */
      if( marks[i*height+j] ) {
				marks[i*height+j] = false; /* all marks become false */
				doClearMark( i, j );
				count++;
      } else {
				k++;
				if( tmp ) occupied++; /* count tiles in use */
      }
    }
    for( j=k; j<height; j++ ) {
      Nfeld[i*height+j] = 0;
    }
  }

  /* Finds and removes empty columns. */
  for( i=0, k=0; i<width; i++ ) {
    for( j=0; j<height; j++ ) {
      Nfeld[k*height+j] = Nfeld[i*height+j];
    }
    if( Nfeld[i*height] /* != 0 */ ) { /* empty bottom means empty column */
      k++;
    }
  }
  /* Clean unfinished empty colums */
  for( i=k; i<width; i++ ) {
    for( j=0; j<height; j++ ) {
      Nfeld[i*height+j] = 0;
    }
  }

  switch( mode ) {
  case modeHighScore:
  case modeLowScore:
    score += (Long)(count-2)*(Long)(count-2);
    break;
  case modeLeaveBoard:
  case modeCleanBoard:
    score = occupied;
    break;
  default:
  }
  doSetScore( score );


  for( i=0; i<width*height; i++ ) {
    if( (tmp=Nfeld[i]) != feld[i] ) {
			feld[i] = tmp;
			doPaint( i/height, i%height );
    }
  }

  /* Check for game over */
  for( i=0; i<width; i++ ) {
    col = i*height;
    for( j=0; j<height; j++ ) {
      if( feld[col+j] ) {
				if( i>0 && feld[col+j]==feld[col+j-height] ) return false;
				if( j>0 && feld[col+j]==feld[col+j-1] ) return false;
      }
    }
  }
  return true;
}

Boolean isClean() {
	Int i;
  for( i=0; i<height*width; i++ ) {
    if( feld[i] ) return false;
	}
	return true;
}

void UnMark() {
  Int i;

  for( i=0; i<height*width; i++ ) {
    if( marks[i] ) {
      marks[i] = false;
      doClearMark( i/height, i%height );
    }
  }
}

void doUndo() {
  Int i;
	UndoData *u;

	if( undoActive ) cheater = true;
  UnMark();

	u = undo[--undoCount];
  doSetScore( score = (u->score&0x3FFFFFFF) );

  for( i=0; i<height*width; i++ ) {
    if( feld[i]!=u->data[i] ) {
      feld[i] = u->data[i];
      doPaint( i/height, i%height );
    }
  }

	MemPtrFree( u->data );
	MemPtrFree( u );
	undo[undoCount] = NULL;
	undoActive = true;
}

Boolean Undo() {
  if( !undoCount ) {
    FrmAlert( NoUndo );
    return false;
  }
  doUndo();
  return true;
}

void Flip() {
  Int i;
  UnMark();
  for( i=0; i<height*width; i++ ) {
    if( feld[i] ) feld[i] = feld[i] % colors + 1;
  }
}

Boolean Empty() {
	Int i;
	// this could be done mch faster by checking the lowest row,
	// but as this function is only called once per game, this not
	// a real performance loss.
	for( i=0; i<height*width; i++ ) {
		if( feld[i] ) return false;
	}
	return true;
}
