/*
 * File: hmaki.c
 * Author: Holger Klawitter
 * Datum: 3.7.1998
 *
 * Changes:
 * 10.06.1999  added support for flipping patterns
 * 18.11.1999  clearing high scores
 * 23.02.2000  color support
 */

//#include <System/SysAll.h>
//#include <UI/UIAll.h>

#include "hmaki.h"
#include "feld.h"
#include "hof.h"
#include "hofDB.h"
#include "util.h"
#include "menu-keys.h"

void DisplayAll( Boolean doCreate, Boolean seeded );

UInt16 osVersion;
Boolean colorDevice;
Boolean bwMode;

/**********************************************************************
 * Global variables
 **********************************************************************/

/***********************************************************************
 * Event Handler for the about form
 ***********************************************************************/
Boolean AboutFormHandleEvent( EventPtr eptr )
{
	Boolean handled = false;
	FormPtr frm = FrmGetActiveForm();

  switch( eptr->eType ) {

  case ctlSelectEvent:
    switch( eptr->data.ctlSelect.controlID ) {
    case AboutOk:
      FrmReturnToForm( MainForm );
			//DisplayAll( false, false );
			handled = true;
    }
    break;

	case frmCloseEvent:
		FrmEraseForm(frm);
		FrmDeleteForm(frm);
		handled = true;
		break;

  case frmOpenEvent:
    FrmDrawForm(frm);
    handled = true;
		break;

  default:
  }

  return handled;
}

/***********************************************************************
 * Event Handler for the seed form
 ***********************************************************************/

Word theSeed = 4711;

Long FieldGetInt( Word id, Long min, Long max, Long def );

Boolean SeedFormHandleEvent( EventPtr eptr )
{
  FormPtr frm = FrmGetActiveForm();
	Boolean handled = false;

  switch( eptr->eType ) {

  case ctlSelectEvent:
    switch( eptr->data.ctlSelect.controlID ) {

    case SeedOk:
      theSeed = FieldGetInt( SeedFld, 0, 99999999, 4711 );
      FrmReturnToForm( MainForm );
      DisplayAll( true, true );
      handled = true;
			break;

    case SeedAbort:
      FrmReturnToForm( MainForm );
			handled = true;
			break;
    }
    break;

	case frmUpdateEvent:
		FrmDrawForm( frm );
		handled = true;
		break;

	case frmCloseEvent:
		FrmEraseForm( frm );
		FrmDeleteForm( frm );
		handled = true;
		break;

  case frmOpenEvent:
    FrmDrawForm( frm );
    FrmSetFocus( frm, FrmGetObjectIndex( frm, SeedFld ) );
		handled = true;
		break;

  default:
  }

  return handled;
}

/***********************************************************************
 * Event Handler for the Hall of Fame form
 ***********************************************************************/

// Contact place for starting a new game when leaving the HOF Form.
Boolean newGameRequest = false;

Boolean HOFFormHandleEvent( EventPtr eptr ) {
  Boolean handled = false;
	FormPtr frm = FrmGetActiveForm();

  switch( eptr->eType ) {
  case ctlSelectEvent:
    switch( eptr->data.ctlSelect.controlID ) {

		case HOFNew:
			newGameRequest = true;
			/* no break: fall though */

		case HOFOk:
      HOFClose();
      FrmGotoForm( MainForm );
			handled = true;
			break;

		case HOFFlush:
			if( FrmAlert( FlushAlert ) ) {
				RectangleType rec = { {0,0},{160,160} };

				HOFClear();
				PutHof( &currentHof );
				dontScore = true;
				WinEraseRectangle( &rec, 0 );
				FrmDrawForm( frm );
				HOFOpen( false, false, score, theSeed );
			}
			handled = true;
			break;
    }
    break;

  case keyDownEvent:
    if( eptr->data.keyDown.chr=='\n' ) {
      HOFClose();
      FrmGotoForm( MainForm );
      handled = true;
    }
    break;

	case frmUpdateEvent:
		FrmDrawForm( frm );
    HOFOpen( cheater, isClean(), score, theSeed );
		handled = true;
		break;

	case frmCloseEvent:
		FrmEraseForm( frm );
		FrmDeleteForm( frm );
		handled = true;
		break;

  case frmOpenEvent:
		if( Empty() /*gameisover*/ ) FrmAlert( GameOverClean );
    FrmDrawForm( frm );
    HOFOpen( cheater, isClean(), score, theSeed );
    handled = true;

  default:
  }
  return handled;
}

/***********************************************************************
 * Event Handler for the OptionsForm
 ***********************************************************************/

/* Prints val into field with id.
 * pre: 0 <= val <= 99!
 */
void FieldSetInt( Word id, UInt val ) {
  FieldPtr fld;
  VoidHand hand;
  CharPtr text;

  fld = GetObjectPtr( id );
  hand = (VoidHand)FldGetTextHandle( fld );
  if( !hand ) {
    //FldSetTextHandle( fld, (VoidPtr)(hand=MemHandleNew(3)) );
    hand = MemHandleNew(3);
  }
  text = MemHandleLock( hand );
	StrIToA( text, val%100 );
  MemHandleUnlock( hand );
  FldSetTextHandle( fld, (VoidPtr)hand );
  FldDrawField( fld );
}

/* Reads an Int from the field.
 * Returns def, if value not in [min,max] min must be >= 0 !
 */
Long FieldGetInt( Word id, Long min, Long max, Long def ) {
  FieldPtr fld;
  CharPtr text;
  Long res = 0;

  fld = GetObjectPtr( id );
  text = FldGetTextPtr( fld );
  if( !text ) return def;
  for( ; *text; text++ ) {
    res = 10*res + *text - '0';
  }
  if( res<min || res>max ) return def;
  return res;
}

void DisplayOptData( Word w, Word h, Word c, Word m ) {
  FieldSetInt( OptWidth, w );
  FieldSetInt( OptHeight, h );
  FieldSetInt( OptColors, c );
  FrmSetControlGroupSelection( FrmGetActiveForm(), 1,
															 m - modeHighScore + OptHighScore ); 
	CtlSetValue( GetObjectPtr( OptBW ), bwMode );
}

Boolean OptFormHandleEvent( EventPtr eptr )
{
  FormPtr frm = FrmGetActiveForm();
  Word aim;
	Boolean handled = false;

  switch( eptr->eType ) {

  case ctlSelectEvent:
    switch( eptr->data.ctlSelect.controlID ) {
			
    case OptOk:
			bwMode = CtlGetValue( GetObjectPtr( OptBW ) );
      currentHof.width = FieldGetInt( OptWidth, MINW, MAXW, DEFW );
      currentHof.height = FieldGetInt( OptHeight, MINH, MAXH, DEFH );
      currentHof.colors = FieldGetInt( OptColors, MINC, MAXC, DEFC );
      aim = FrmGetObjectId( frm, FrmGetControlGroupSelection( frm, 1 ) );
      currentHof.mode = aim - OptHighScore + modeHighScore;
      if( !GetHof( currentHof.width,
									 currentHof.height,
									 currentHof.colors,
									 currentHof.mode, &currentHof ) ) {
				HOFClear();
				PutHof( &currentHof );
      }
      FrmReturnToForm( MainForm );
      DisplayAll( true, false );
			handled = true;
			break;

    case OptCancel:
      FrmReturnToForm( MainForm );
			handled = true;
			break;

		case OptPurge:
			if( FrmAlert( PurgeAlert ) ) {
				PurgeHOF();
				HOFClear();
				PutHof( &currentHof );
			}
			handled = true;
			break;

    default:
    }
    break;

	case frmCloseEvent:
		FrmEraseForm( frm );
		FrmDeleteForm( frm );
		handled = true;
		break;

  case frmOpenEvent:
    FrmDrawForm( frm );
    DisplayOptData( currentHof.width, currentHof.height, currentHof.colors,
										currentHof.mode );
		handled = true;
		break;

  default:
  }

  return handled;
}

/***********************************************************************
 * Event Handler for the MainForm
 ***********************************************************************/

Boolean gameIsOver = false;

Char finalscore[11];
/* String rep of score at end of game, if game was aimed for high score.
 */

void DisplayAll( Boolean doCreate, Boolean seeded ) {
  FormPtr frm = FrmGetActiveForm();
  Word cleanIdx = FrmGetObjectIndex( frm, MakiCleanBoard );
  Word highIdx = FrmGetObjectIndex( frm, MakiHighScore );

  if( doCreate ) {
    if( !seeded ) {
      theSeed = SysRandom(0);
    }
    Create(theSeed);
    gameIsOver = false;
  }
  FrmDrawForm(FrmGetActiveForm());
  doSetScore( score );
  switch( currentHof.mode ) {
  case modeHighScore:
  case modeLowScore:
    FrmHideObject( frm, cleanIdx );
    FrmShowObject( frm, highIdx );
    break;
  case modeCleanBoard:
  case modeLeaveBoard:
    FrmHideObject( frm, highIdx );
    FrmShowObject( frm, cleanIdx );
  }
  Paint();
}

Boolean wantsToDrop = false;

Boolean MainFormHandleEvent( EventPtr eptr )
{
  MenuBarPtr menu;
  Char seedStr[11];
	FormPtr frm = FrmGetActiveForm();
	Boolean handled = false;

  switch( eptr->eType ) {

  case menuEvent:
    menu = MenuGetActiveMenu();
    if( menu ) MenuEraseStatus( menu );
    switch( eptr->data.menu.itemID ) {

    case MenuAbout:
      FrmPopupForm( AboutForm );
			handled = true;
      break;

    case MenuNew:
      DisplayAll( true, false );
			handled = true;
      break;

    case MenuSeed:
      FrmPopupForm( SeedForm );
			handled = true;
      break;

    case MenuOpt:
      FrmPopupForm( OptForm );
			handled = true;
      break;

    case MenuUndo:
      Undo();
			handled = true;
      break;

    case MenuHOF:
      dontScore = true;
      FrmGotoForm( HOFForm );
			handled = true;
      break;

    case MenuShSeed:
      StrIToA( seedStr, theSeed );
      FrmCustomAlert( ShowSeed, seedStr, NULL, NULL );
			handled = true;
      break;

    case MenuFlip:
      Flip();
			DisplayAll( false, false );
			handled = true;
      break;

    default:
    }
		break;

  case keyDownEvent:
    switch( eptr->data.keyDown.chr ) {

    case Menu_about:
      FrmPopupForm( AboutForm );
			handled = true;
			break;

    case Menu_seed:
      FrmPopupForm( SeedForm );
			handled = true;
			break;

    case Menu_new:
      DisplayAll( true, false );
			handled = true;
			break;

    case Menu_prefs:
      FrmPopupForm( OptForm );
			handled = true;
			break;

		case Menu_flip:
			Flip();
			DisplayAll( false, false );
			handled = true;
			break;

    case Menu_undo:
      Undo();
			handled = true;
			break;

    case Menu_hof:
      dontScore = true;
      FrmGotoForm( HOFForm );
			handled = true;
			break;

		case Menu_shseed:
      StrIToA( seedStr, theSeed );
      FrmCustomAlert( ShowSeed, seedStr, NULL, NULL );
			handled = true;
			break;

    }
		break;

  case penDownEvent:
    if ( gameIsOver && isLegal( eptr->screenX/hsize,
																(160-eptr->screenY)/vsize ) ) {
      if( FrmAlert( NewGame ) == 0 ) {
				DisplayAll( true, false );
      }
    } else {
      wantsToDrop = Mark( eptr->screenX/hsize, (160-eptr->screenY)/vsize );
    }
		/* handled = false */
		break;
		
  case penUpEvent:
    if( isMarked( eptr->screenX/hsize, (160-eptr->screenY)/vsize ) ) {
      if( wantsToDrop && Drop() ) {
				gameIsOver = true;
				FrmGotoForm( HOFForm );
				wantsToDrop = false;
      } /* if( .. && Drop() ) */
    } else {
      UnMark();
    }
		handled = true;
		break;

	case frmUpdateEvent:
		FrmDrawForm( frm );
		DisplayAll( false, false );
		handled = true;
		break;

	case frmCloseEvent:
		FrmEraseForm( frm );
		FrmDeleteForm( frm );
		handled = true;
		break;

  case frmOpenEvent:
    DisplayAll( newGameRequest, false );
		newGameRequest = false;
		handled = true;
		break;

  default:
  }

  return handled;
}

/***********************************************************************
 * MAIN PROGRAM STUFF
 ***********************************************************************/

typedef struct {
  Word width, height, colors, aim;
  Byte board[MAXW*MAXH];
  Long score;
  Boolean gameIsOver;
  Word seed;
	Boolean cheater;
	Boolean bw;
} MakiPrefs;

/***********************************************************************
 * Starting and ending the application
 **********************************************************************/

static Word StartApplication()
{
  MakiPrefs prefs;
	UInt32 romVersion;
	DWord dummy;

	FtrGet( sysFtrCreator, sysFtrNumROMVersion, &romVersion );
	osVersion = romVersion >> 20;
	colorDevice = false;
	if( osVersion >= 0x035 ) {
		WinScreenMode( winScreenModeGetDefaults,
									 &dummy, &dummy, &dummy, &colorDevice );
		//if( colorDevice ) { DebugMsg( "Color Device!" ); }
	}

  OpenHofDB();
  if( PrefGetAppPreferencesV10( MakiID, MakiVersion,
																&prefs, sizeof(MakiPrefs) )
      && GetHof( prefs.width, prefs.height, prefs.colors, prefs.aim,
								 &currentHof )
			) {
    score = prefs.score;
    gameIsOver = prefs.gameIsOver;
    theSeed = prefs.seed;
		bwMode = prefs.bw;
    Alloc();
    MemMove( feld, prefs.board, prefs.width*prefs.height );
		cheater = prefs.cheater;
  } else {
    currentHof.width = DEFW;
    currentHof.height = DEFH;
    currentHof.colors = DEFC;
    currentHof.mode = DEFM;
    HOFClear();
    Create(theSeed=SysRandom(0));
    gameIsOver = false;
  }
  return 0;
}

static void StopApplication()
{
  MakiPrefs prefs;
  prefs.width = currentHof.width;
  prefs.height = currentHof.height;
  prefs.colors = currentHof.colors;
  prefs.aim = currentHof.mode;
  prefs.score = score;
  prefs.gameIsOver = gameIsOver;
  prefs.seed = theSeed;
	prefs.cheater = cheater;
	prefs.bw = bwMode;
  MemMove( prefs.board, feld, prefs.width*prefs.height );
  PrefSetAppPreferencesV10( MakiID, MakiVersion, &prefs, sizeof(MakiPrefs) );
  PutHof( &currentHof );
  CloseHofDB();
	Dealloc();
}

/***********************************************************************
 * The master Event handler stuff
 ***********************************************************************/
static Boolean ApplicationHandleEvent( EventPtr eptr )
{
  FormPtr frm;
  Int formId;
  Boolean handled = false;

  if( eptr->eType == frmLoadEvent ) {
    formId = eptr->data.frmLoad.formID;

    frm = FrmInitForm( formId );
    FrmSetActiveForm( frm );

    switch( formId ) {
    case MainForm:
      FrmSetEventHandler( frm, (MainFormHandleEvent) );
      break;
    case OptForm:
      FrmSetEventHandler( frm, (OptFormHandleEvent) );
      break;
    case AboutForm:
      FrmSetEventHandler( frm, (AboutFormHandleEvent) );
      break;
    case HOFForm:
      FrmSetEventHandler( frm, (HOFFormHandleEvent) );
      break;
    case SeedForm:
      FrmSetEventHandler( frm, (SeedFormHandleEvent) );
      break;
    }
    handled = true;
  }  
  return handled;
}

static void EventLoop()
{
  EventType e;
  Word err;

  do {
    EvtGetEvent( &e, evtWaitForever );
    if( SysHandleEvent( &e ) ) continue;
		if( MenuHandleEvent( NULL, &e, &err ) ) continue;
		if( ApplicationHandleEvent( &e ) ) continue;
		FrmDispatchEvent( &e );
  } while( e.eType != appStopEvent );
}

/***********************************************************************
 * The Main loop
 ***********************************************************************/
DWord PilotMain(Word cmd, Ptr cmdPBP, Word launchFlags)
{
  Word err;

  if( cmd == sysAppLaunchCmdNormalLaunch ) {
    err = StartApplication();
    if( err ) {
      return err;
    }
    FrmGotoForm( MainForm );
    EventLoop();
    StopApplication();
		FrmCloseAllForms();
  }

  return 0;
}
