/*
 * Datei: hof.c
 * Autor: Holger Klawitter
 * Datum: 13.8.1998
 *
 * Changes:
 * 27.09.1999 HK multilingual support - strings > resources
 */
#include "hof.h"
#include "hofDB.h"
#include "hmaki.h"
#include "feld.h"
#include <PalmOS.h>
#include <DLServer.h>
#include "hmaki.rch"

/**********************************************************************
 * Global variables
 **********************************************************************/

Hof currentHof;

Boolean dontScore = false;

UInt modeStringMsgID[4] = {
	strModeBigTiles,
	strModeCleanUp,
	strModeLeaveBoard,
	strModeSmallTiles
};

/**********************************************************************
 * Functions
 **********************************************************************/

void HOFClear() {
  Int i;
  Long defaultScore;

  switch( currentHof.mode ) {
  case modeHighScore:
  case modeLeaveBoard:
    defaultScore = 0;
    break;
  case modeCleanBoard: 
    defaultScore = currentHof.width * currentHof.height;
    break;
  case modeLowScore: 
  default:
    defaultScore = currentHof.width * currentHof.height - 2;
    defaultScore *= defaultScore;
    break;
  }

  for( i=0; i<SCORES; i++ ) {
    StrCopy( currentHof.tbl[i].name, "HMaki" );
    currentHof.tbl[i].score = defaultScore;
    currentHof.tbl[i].seed = 0;
  }
  PutHof( &currentHof );
}

Int pos = SCORES;
Word enterID;

/* Hold den String mit der id aus der Resourcedatenbank und
 * hngt ihn an den stringpointer an.
 */
void StringAppend( CharPtr str, UInt rsrcId ) {
	VoidHand res = DmGet1Resource( 'tSTR', rsrcId );
	CharPtr resStr;

	ErrFatalDisplayIf( !res, "string not found" );
	resStr = MemHandleLock( res );
  StrCat( str, resStr );
	MemHandleUnlock( res );
	DmReleaseResource( res );
}

void HOFOpen(Boolean cheated,Boolean cleaned,Long newScore,Word newSeed) {
  Int i,x;
  FormPtr frm;
  FieldPtr fld;
  Char str[30];
	Char modeLine[40] = "";
  CharPtr txt,s;
	Char userName[dlkUserNameBufSize];
  FontID tmp;
	Long lscore;

  VoidHand hand;

  tmp = FntSetFont( boldFont );

  if( dontScore ) {
    pos=SCORES;
  } else {
    /* find insertion pos */
    for( pos=0; pos<SCORES; pos++ ) {
			lscore = (currentHof.tbl[pos].score & 0x3FFFFFFF);
      switch( currentHof.mode ) {
      case modeHighScore:
      case modeLeaveBoard:
				if( lscore<newScore ) goto breakFor;
				break;
      case modeCleanBoard:
      case modeLowScore:
				if( lscore>newScore ) goto breakFor;
      }
    }
  breakFor:
  }
	
  /* No insertion point found => pos==SCORES.
   * We will use that a couple of times!
   */

  /* Make space */
  for( i=SCORES-1; i>pos; i-- ) {
    MemMove( &(currentHof.tbl[i]), &(currentHof.tbl[i-1]), sizeof(HofEntry) );
  }
  /*
    MemMove( currentHof.tbl[i], currentHof.tbl[i-1], sizeof(HofEntry)
    *(SCORES-pos) );
		*/

  if( pos<SCORES ) {
    currentHof.tbl[pos].score = newScore;
		if( cheated ) currentHof.tbl[pos].score |= 0x80000000;
		if( cleaned ) currentHof.tbl[pos].score |= 0x40000000;
    currentHof.tbl[pos].seed = newSeed;

    enterID = HOFEnter+pos;
    fld = GetObjectPtr( enterID );
		hand = FldGetTextHandle( fld );
		if( !hand ) {
			hand = MemHandleNew( (NAMELEN+1)*sizeof(Char) );
		} else if( MemHandleSize(hand)<NAMELEN+1 ) {
			MemHandleResize( hand, NAMELEN+1 );
		}
    txt = MemHandleLock( hand );
		DlkGetSyncInfo(0,0,0,userName,0,0);
		StrNCopy( txt, userName, NAMELEN );
		txt[NAMELEN] = '\0';
		s = StrChr( txt, ' ' );
		if( s ) *s = '\0';
    FldSetTextHandle( fld, (Handle)hand );
    MemHandleUnlock( hand );

    FldSetUsable( fld, true );
    FldDrawField( fld );
    frm = FrmGetActiveForm();
    FrmSetFocus( frm, FrmGetObjectIndex(frm,enterID) );
		FldSetSelection( fld, 0, StrLen(txt) );
  }

	StringAppend( modeLine, strGameLabelID );

  StrIToA( modeLine+StrLen(modeLine), currentHof.width );
  StrCat( modeLine, "x" );
  StrIToA( modeLine+StrLen(modeLine), currentHof.height );
  StrCat( modeLine, "x" ); 
  StrIToA( modeLine+StrLen(modeLine), currentHof.colors );

  switch( currentHof.mode ) {
  case modeCleanBoard:
  case modeLeaveBoard:
  case modeHighScore:
  case modeLowScore:
		StringAppend( modeLine, modeStringMsgID[currentHof.mode - modeHighScore] );
    break;
  }
  x = 78 - FntCharsWidth( modeLine, i=StrLen(modeLine) ) / 2;
  WinDrawChars( modeLine, i, x, 18 );
 
  /* Now the display */
  for( i=0; i<SCORES; i++ ) {
		lscore = currentHof.tbl[i].score;

    FntSetFont( boldFont );

		// Name
    if( i!=pos ) {
      txt = currentHof.tbl[i].name;
      WinDrawChars( txt, StrLen(txt), 15, 35+i*20 );
    }

		// Score
    StrIToA( str, (lscore & 0x3fffffff) );
    x = 115 - FntCharsWidth( str, StrLen(str) );
    WinDrawChars( str, StrLen(str), x, 35+i*20 );

    FntSetFont( stdFont );

		// Seed
    StrIToA( str, currentHof.tbl[i].seed );
    x = 150 - FntCharsWidth( str, StrLen(str) );
    WinDrawChars( str, StrLen(str), x, 35+i*20 );

		// Flags
		StrCopy( str, "" );
		if( (lscore & 0x80000000)!=0 ) {
			StringAppend( str, strCheated );
		}
		if( (lscore & 0x40000000)!=0 ) {
			StringAppend( str, strClean );
		}
		if( StrLen(str)>0 ) {
			x = 150 - FntCharsWidth( str, StrLen(str) );
			WinDrawChars( str, StrLen(str), x, 44+i*20 );
		}
  }
  FntSetFont( tmp );
}

void HOFClose() {
  FieldPtr fld;
  CharPtr name;

  if( pos!=SCORES ) {
    fld = GetObjectPtr( enterID );
    name = FldGetTextPtr( fld );
    MemSet( currentHof.tbl[pos].name, 0, sizeof(Char)*NAMELEN );
    if( name ) {
      StrCopy( currentHof.tbl[pos].name, name );
    }
    FldSetUsable( fld, false );
  }
  PutHof( &currentHof );
  dontScore = false;
}

/**********************************************************************
 * private functions
 **********************************************************************/
