//#include <PalmTypes.h>
#include "Compat.h"
#include <PalmOS.h>

//#include <callback.h>

#include "hofDB.h"

static DmOpenRef hofDB;

/*********** PROTOTYPES ***********/

static Int HofComp( Hof* p1, Hof* p2, Int dummy,
		    void* d1, void* d2, void* d3 );

/*********** IMPLEMENTATION ***********/

/* Datenbank ffnen und ggf. anlegen */
Boolean OpenHofDB() {
  Int err;
  KillOldHofDB();
  hofDB = DmOpenDatabaseByTypeCreator( HofDBType, HofDBID, dmModeReadWrite );
  if( hofDB ) return false;
  err = DmCreateDatabase( 0, HofDBName, HofDBID, HofDBType, false );
  if( err ) return true;
  hofDB = DmOpenDatabaseByTypeCreator( HofDBType, HofDBID, dmModeReadWrite );
  return false;
}

/* Datenbank schliessen */
void CloseHofDB() {
  DmCloseDatabase( hofDB );
}

/* Lschen der kompletten DB, das sollte die mit der ID HK02 sein */
void KillOldHofDB() {
  LocalID dbID;
  UInt card;
  Err res;
  Char name[32];
  DmSearchStateType dummy;

  res = DmGetNextDatabaseByTypeCreator
    ( true, &dummy, HofDBType, 'HK02', false, &card, &dbID );
  if( res == 0 ) {
    res = DmDatabaseInfo( card, dbID, name,
	NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL, NULL );
    if( res == 0 && StrCompare( name, HofDBName ) == 0 ) {
      DmDeleteDatabase( card, dbID );
    }
  }
}

/* Lschen der aktuellen DB und gleich wieder Anlegen einer neuen. */
void PurgeHOF() {
	Err res;
	DmSearchStateType dummy;
  LocalID dbID;
  UInt card;

	CloseHofDB();
	res = DmGetNextDatabaseByTypeCreator
    ( true, &dummy, HofDBType, HofDBID, false, &card, &dbID );
	if( res ) {
		DmDeleteDatabase( card, dbID );
	}
	OpenHofDB();
}

/* Hof mit dem Namen aus der Datenbank auslesen,
 * gibt false zurck, wenn es den Hof nicht gibt. */
Int GetHof( Word w, Word h, Word c, Word m, Hof* hof ) {
  Hof dummy;
  Hof *hofFromDB;
  UInt idx;
  VoidHand hofHandle;
  Boolean res = false;

  dummy.width = w;
  dummy.height = h;
  dummy.colors = c;
  dummy.mode = m;

  idx = DmFindSortPositionV10( hofDB, &dummy, (DmComparF*)HofComp, 0 );
  if( idx > DmNumRecords( hofDB ) || idx == 0 ) return false;
  /* Da ein neuer Eintrag mit demselben Namen hinter diesem eingefgt
   * werden wrde, mu man einen Schritt zurckgehen.
   */
  idx--;
  hofHandle = DmGetRecord( hofDB, idx );
  hofFromDB = MemHandleLock( hofHandle );
  if( HofComp( hofFromDB, &dummy, 0,0,0,0 )==0 ) {
    MemMove( hof, hofFromDB, sizeof(Hof) );
    res = true;
  }
  MemHandleUnlock( hofHandle );
  DmReleaseRecord( hofDB, idx, false );
  return res;
}

void GetHofByIdx( Int idx, Hof* hof ) {
  VoidHand handle = DmGetRecord( hofDB, idx );
  Hof *hofFromDB = MemHandleLock( handle );
  MemMove( hof, hofFromDB, sizeof(Hof) );
  MemHandleUnlock( handle );
  DmReleaseRecord( hofDB, idx, false );
}

/* Hof in die Datenbank einfgen.
 * Ist eine Hof mit dem selben Namen vorhanden, wird diese
 * berschrieben.
 */
void PutHof( Hof* hof ) {
  Hof *hofFromDB;
  UInt idx;
  Err err;
  VoidHand hofHandle = NULL;
  Boolean newHof = true;

  idx = DmFindSortPositionV10( hofDB, hof, (DmComparF*)HofComp, 0 );
  if( idx > 0 ) {
    idx--;
    hofHandle = DmGetRecord( hofDB, idx );
    hofFromDB = MemHandleLock( hofHandle );
    if( HofComp( hof, hofFromDB, 0,0,0,0 )==0 ) {
      /* Bestehender Eintrag wird verndert! */
      err = DmWrite( hofFromDB, 0, hof, sizeof(Hof) );
      newHof = false;
    } 
    MemHandleUnlock( hofHandle );
    DmReleaseRecord( hofDB, idx, false );
    /* Falls doch eine neue Hof eigefgt werden soll, mu der Index
     * wieder hergestellt werden. Ansonsten ist der neue Wert egal.
     */
    idx++;
  }

  if( newHof ) {
    hofHandle = DmNewRecord( hofDB, &idx, sizeof(Hof) );
    hofFromDB = MemHandleLock( hofHandle );
    err = DmWrite( hofFromDB, 0, hof, sizeof(Hof) );
    MemHandleUnlock( hofHandle );
    DmReleaseRecord( hofDB, idx, false );
  }
}

/* Hof aus der Datenbank lschen.
 * Wenn es keine Eintrag dieses Namens gibt, tue nichts.
 */
void DelHof( Word w, Word h, Word c, Word m ) {
  Hof dummy = { w, h, c, m, {
    {{NULL},0},
    {{NULL},0},
    {{NULL},0},
    {{NULL},0},
    {{NULL},0},
  }};
  Hof *hofFromDB;
  VoidHand hofHandle;
  UInt idx;
  Boolean killIt = false;

  idx = DmFindSortPositionV10( hofDB, &dummy, (DmComparF*)HofComp, 0 );
  if( idx > 0 ) {
    idx--;
    hofHandle = DmGetRecord( hofDB, idx );
    hofFromDB = MemHandleLock( hofHandle );
    if( HofComp( hofFromDB, &dummy, 0,0,0,0 )==0 ) {
      killIt = true;
    }
    MemHandleUnlock( hofHandle );
    DmReleaseRecord( hofDB, idx, false );
  }

  if( killIt ) {
    DmRemoveRecord( hofDB, idx );
  }
}

static Int HofComp( Hof* p1, Hof* p2, Int dummy,
		       void* d1, void* d2, void* d3 ) {
  Int res;
  //CALLBACK_PROLOGUE;
       if( p1->width > p2->width ) res = 1;
  else if( p1->width < p2->width ) res = -1;
  else if( p1->height > p2->height ) res = 1;
  else if( p1->height < p2->height ) res = -1;
  else if( p1->colors > p2->colors ) res = 1;
  else if( p1->colors < p2->colors ) res = -1;
  else if( p1->mode > p2->mode ) res = 1;
  else if( p1->mode < p2->mode ) res = -1;
  else res = 0;
       //CALLBACK_EPILOGUE;
  return res;
}

Int GetHofListSize() {
  return DmNumRecords( hofDB );
}
